`include "timescale.v"
`include "aes_regdef.v"

module aes_h8_tb;

   reg CLK;
   reg nRST;
   reg nCS;
   reg nWE;
   reg nRD;
   reg nAS;
   reg [6:0] ADDR;
   reg [7:0] DATA_IN;
   wire [7:0] DATA_OUT;
   wire [7:0] DATA;
   assign     DATA = (nWE == 0 ? DATA_IN : 8'bz);
   assign     DATA_OUT = (nRD == 0 ? DATA : 8'bz);

   reg [127:0] key;
   reg [127:0] iv;
   reg [127:0] plain[3:0];
   reg [127:0] cipher[3:0];
   reg [127:0] aes_in;
   reg [127:0] aes_out;
   integer      i;
   reg [7:0] 	status;
   reg [127:0] 	iv_tmp;

   // AES H8
   aes_h8 aes_h8(.CLK(CLK), .nRST(nRST),
		 .nCS(nCS), .nWE(nWE), .nOE(nRD),
		 .ADDR(ADDR), .DATA(DATA), .nAS(nAS));

   // Write task (3 states access @ 25MHz)
   task write;
      input [6:0] addr;
      input [7:0] d;
      begin
	 @(posedge CLK);
	 #25 ADDR = addr; nCS = 0;
	 #30 DATA_IN = d;
	 #10 nAS = 0; nWE = 0;
	 #35 nWE = 1;
	 #30 nAS = 1; nCS = 1; DATA_IN = 8'bz;
	 #20;
      end
   endtask

   // Read task (3 states access @ 25MHz)
   task read;
      input [6:0] addr;
      output [7:0] d;
      begin
	 @(posedge CLK);
	 #25 ADDR = addr; nCS = 0;
	 #25 nAS = 0; nRD = 0;
	 #55 d = DATA_OUT;
	 #25 nAS = 1; nCS = 1;
	 #25 nRD = 1; 
	 #5;
        end
   endtask

   // Clock
   always
     begin
	CLK = 1'b0;
	#5 CLK = 1'b1;
	#20 CLK = 1'b0;
	#15 CLK = 1'b0;
     end
   
   initial begin

      key = 128'h2b7e151628aed2a6abf7158809cf4f3c;
      iv = 128'h000102030405060708090a0b0c0d0e0f;

      plain[0] = 128'h6bc1bee22e409f96e93d7e117393172a;
      cipher[0] = 128'h7649abac8119b246cee98e9b12e9197d;
      plain[1] = 128'hae2d8a571e03ac9c9eb76fac45af8e51;
      cipher[1] = 128'h5086cb9b507219ee95db113a917678b2;
      plain[2] = 128'h30c81c46a35ce411e5fbc1191a0a52ef;
      cipher[2] = 128'h73bed6b8e3c1743b7116e69e22229516;
      plain[3] = 128'hf69f2445df4f9b17ad2b417be66c3710;
      cipher[3] = 128'h3ff1caa1681fac09120eca307586e1a7;

      // --------------------
      nRST = 1'b1;
      nCS = 1'b1;
      nWE = 1'b1;
      nRD = 1'b1;
      ADDR = 6'b000000; //0
      nAS = 1'b1;
      // --------------------
      #50 // Time=50 ns
	nRST = 1'b0;
      // --------------------
      #25 // Time=75 ns
	// --------------------
	nRST = 1'b1;

      @(posedge CLK);
      @(posedge CLK);
      @(posedge CLK);
      @(posedge CLK);

      // Test Vector from
      // http://csrc.nist.gov/publications/nistpubs/800-38a/sp800-38a.pdf
      //
      // F.2.1 CBC-AES128.Encrypt
      // Key		2b7e151628aed2a6abf7158809cf4f3c
      // IV		000102030405060708090a0b0c0d0e0f
      //
      // Block #1
      // Plaintext	6bc1bee22e409f96e93d7e117393172a
      // Input Block	6bc0bce12a459991e134741a7f9e1925
      // Output Block	7649abac8119b246cee98e9b12e9197d
      // Ciphertext	7649abac8119b246cee98e9b12e9197d
      //
      // Block #2
      // Plaintext	ae2d8a571e03ac9c9eb76fac45af8e51
      // Input Block	d86421fb9f1a1eda505ee1375746972c
      // Output Block	5086cb9b507219ee95db113a917678b2
      // Ciphertext	5086cb9b507219ee95db113a917678b2
      //
      // Block #3
      // Plaintext	30c81c46a35ce411e5fbc1191a0a52ef
      // Input Block	604ed7ddf32efdff7020d0238b7c2a5d
      // Output Block	73bed6b8e3c1743b7116e69e22229516
      // Ciphertext	73bed6b8e3c1743b7116e69e22229516
      //
      // Block #4
      // Plaintext	f69f2445df4f9b17ad2b417be66c3710
      // Input Block	8521f2fd3c8eef2cdc3da7e5c44ea206
      // Output Block	3ff1caa1681fac09120eca307586e1a7
      // Ciphertext	3ff1caa1681fac09120eca307586e1a7

      // input key
      write(`AES_KEY_0, key[127:120]);
      write(`AES_KEY_1, key[119:112]);
      write(`AES_KEY_2, key[111:104]);
      write(`AES_KEY_3, key[103:96]);
      write(`AES_KEY_4, key[95:88]);
      write(`AES_KEY_5, key[87:80]);
      write(`AES_KEY_6, key[79:72]);
      write(`AES_KEY_7, key[71:64]);
      write(`AES_KEY_8, key[63:56]);
      write(`AES_KEY_9, key[55:48]);
      write(`AES_KEY_10, key[47:40]);
      write(`AES_KEY_11, key[39:32]);
      write(`AES_KEY_12, key[31:24]);
      write(`AES_KEY_13, key[23:16]);
      write(`AES_KEY_14, key[15:8]);
      write(`AES_KEY_15, key[7:0]);

      iv_tmp = iv;

      for (i = 0; i < 4; i = i + 1) begin

	 // input IV
	 write(`AES_IV_0, iv_tmp[127:120]);
	 write(`AES_IV_1, iv_tmp[119:112]);
	 write(`AES_IV_2, iv_tmp[111:104]);
	 write(`AES_IV_3, iv_tmp[103:96]);
	 write(`AES_IV_4, iv_tmp[95:88]);
	 write(`AES_IV_5, iv_tmp[87:80]);
	 write(`AES_IV_6, iv_tmp[79:72]);
	 write(`AES_IV_7, iv_tmp[71:64]);
	 write(`AES_IV_8, iv_tmp[63:56]);
	 write(`AES_IV_9, iv_tmp[55:48]);
	 write(`AES_IV_10, iv_tmp[47:40]);
	 write(`AES_IV_11, iv_tmp[39:32]);
	 write(`AES_IV_12, iv_tmp[31:24]);
	 write(`AES_IV_13, iv_tmp[23:16]);
	 write(`AES_IV_14, iv_tmp[15:8]);
	 write(`AES_IV_15, iv_tmp[7:0]);

	 // input data
	 aes_in = plain[i];
	 write(`AES_DIN_0, aes_in[127:120]);
	 write(`AES_DIN_1, aes_in[119:112]);
	 write(`AES_DIN_2, aes_in[111:104]);
	 write(`AES_DIN_3, aes_in[103:96]);
	 write(`AES_DIN_4, aes_in[95:88]);
	 write(`AES_DIN_5, aes_in[87:80]);
	 write(`AES_DIN_6, aes_in[79:72]);
	 write(`AES_DIN_7, aes_in[71:64]);
	 write(`AES_DIN_8, aes_in[63:56]);
	 write(`AES_DIN_9, aes_in[55:48]);
	 write(`AES_DIN_10, aes_in[47:40]);
	 write(`AES_DIN_11, aes_in[39:32]);
	 write(`AES_DIN_12, aes_in[31:24]);
	 write(`AES_DIN_13, aes_in[23:16]);
	 write(`AES_DIN_14, aes_in[15:8]);
	 write(`AES_DIN_15, aes_in[7:0]);

	 // CBC mode and start trigger
	 write(`AES_CTRL, `AES_CTRL_ENC | `AES_CTRL_CBC | `AES_CTRL_START);

	 // XXX: need time for BUSY bit to be asserted
	 @(posedge CLK);
	 @(posedge CLK);

	 // wait for ...
	 read(`AES_STATUS, status);
	 while (status & `AES_STATUS_BUSY)
	   read(`AES_STATUS, status);

	 // output data
	 read(`AES_DOUT_0, aes_out[127:120]);
	 read(`AES_DOUT_1, aes_out[119:112]);
	 read(`AES_DOUT_2, aes_out[111:104]);
	 read(`AES_DOUT_3, aes_out[103:96]);
	 read(`AES_DOUT_4, aes_out[95:88]);
	 read(`AES_DOUT_5, aes_out[87:80]);
	 read(`AES_DOUT_6, aes_out[79:72]);
	 read(`AES_DOUT_7, aes_out[71:64]);
	 read(`AES_DOUT_8, aes_out[63:56]);
	 read(`AES_DOUT_9, aes_out[55:48]);
	 read(`AES_DOUT_10, aes_out[47:40]);
	 read(`AES_DOUT_11, aes_out[39:32]);
	 read(`AES_DOUT_12, aes_out[31:24]);
	 read(`AES_DOUT_13, aes_out[23:16]);
	 read(`AES_DOUT_14, aes_out[15:8]);
	 read(`AES_DOUT_15, aes_out[7:0]);

	 @(posedge CLK);
	 @(posedge CLK);
	 @(posedge CLK);
	 @(posedge CLK);

	 if (aes_out !== cipher[i])
	   $display("ERROR: Expected %x Got %x", cipher[i], aes_out);
	 else
	   $display("PASS : Expected %x Got %x", cipher[i], aes_out);

	 iv_tmp = aes_out;

      end // for (i = 0; i < 4; i = i + 1)

      //
      // F.2.2 CBC-AES128.Decrypt
      // Key		2b7e151628aed2a6abf7158809cf4f3c
      // IV		000102030405060708090a0b0c0d0e0f
      //
      // Block #1
      // Ciphertext	7649abac8119b246cee98e9b12e9197d
      // Input Block	7649abac8119b246cee98e9b12e9197d
      // Output Block	6bc0bce12a459991e134741a7f9e1925
      // Plaintext	6bc1bee22e409f96e93d7e117393172a
      //
      // Block #2
      // Ciphertext	5086cb9b507219ee95db113a917678b2
      // Input Block	5086cb9b507219ee95db113a917678b2
      // Output Block	d86421fb9f1a1eda505ee1375746972c
      // Plaintext	ae2d8a571e03ac9c9eb76fac45af8e51
      //
      // Block #3
      // Ciphertext	73bed6b8e3c1743b7116e69e22229516
      // Input Block	73bed6b8e3c1743b7116e69e22229516
      // Output Block	604ed7ddf32efdff7020d0238b7c2a5d
      // Plaintext	30c81c46a35ce411e5fbc1191a0a52ef
      //
      // Block #4
      // Ciphertext	3ff1caa1681fac09120eca307586e1a7
      // Input Block	3ff1caa1681fac09120eca307586e1a7
      // Output Block	8521f2fd3c8eef2cdc3da7e5c44ea206
      // Plaintext	f69f2445df4f9b17ad2b417be66c3710

      // input key
      write(`AES_KEY_0, key[127:120]);
      write(`AES_KEY_1, key[119:112]);
      write(`AES_KEY_2, key[111:104]);
      write(`AES_KEY_3, key[103:96]);
      write(`AES_KEY_4, key[95:88]);
      write(`AES_KEY_5, key[87:80]);
      write(`AES_KEY_6, key[79:72]);
      write(`AES_KEY_7, key[71:64]);
      write(`AES_KEY_8, key[63:56]);
      write(`AES_KEY_9, key[55:48]);
      write(`AES_KEY_10, key[47:40]);
      write(`AES_KEY_11, key[39:32]);
      write(`AES_KEY_12, key[31:24]);
      write(`AES_KEY_13, key[23:16]);
      write(`AES_KEY_14, key[15:8]);
      write(`AES_KEY_15, key[7:0]);

      iv_tmp = iv;

      for (i = 0; i < 4; i = i + 1) begin

	 // input IV
	 write(`AES_IV_0, iv_tmp[127:120]);
	 write(`AES_IV_1, iv_tmp[119:112]);
	 write(`AES_IV_2, iv_tmp[111:104]);
	 write(`AES_IV_3, iv_tmp[103:96]);
	 write(`AES_IV_4, iv_tmp[95:88]);
	 write(`AES_IV_5, iv_tmp[87:80]);
	 write(`AES_IV_6, iv_tmp[79:72]);
	 write(`AES_IV_7, iv_tmp[71:64]);
	 write(`AES_IV_8, iv_tmp[63:56]);
	 write(`AES_IV_9, iv_tmp[55:48]);
	 write(`AES_IV_10, iv_tmp[47:40]);
	 write(`AES_IV_11, iv_tmp[39:32]);
	 write(`AES_IV_12, iv_tmp[31:24]);
	 write(`AES_IV_13, iv_tmp[23:16]);
	 write(`AES_IV_14, iv_tmp[15:8]);
	 write(`AES_IV_15, iv_tmp[7:0]);

	 // input data
	 aes_in = cipher[i];
	 write(`AES_DIN_0, aes_in[127:120]);
	 write(`AES_DIN_1, aes_in[119:112]);
	 write(`AES_DIN_2, aes_in[111:104]);
	 write(`AES_DIN_3, aes_in[103:96]);
	 write(`AES_DIN_4, aes_in[95:88]);
	 write(`AES_DIN_5, aes_in[87:80]);
	 write(`AES_DIN_6, aes_in[79:72]);
	 write(`AES_DIN_7, aes_in[71:64]);
	 write(`AES_DIN_8, aes_in[63:56]);
	 write(`AES_DIN_9, aes_in[55:48]);
	 write(`AES_DIN_10, aes_in[47:40]);
	 write(`AES_DIN_11, aes_in[39:32]);
	 write(`AES_DIN_12, aes_in[31:24]);
	 write(`AES_DIN_13, aes_in[23:16]);
	 write(`AES_DIN_14, aes_in[15:8]);
	 write(`AES_DIN_15, aes_in[7:0]);

	 // CBC mode and start trigger
	 write(`AES_CTRL, `AES_CTRL_DEC | `AES_CTRL_CBC | `AES_CTRL_START);

	 // XXX: need time for BUSY bit to be asserted
	 @(posedge CLK);
	 @(posedge CLK);

	 // wait for ...
	 read(`AES_STATUS, status);
	 while (status & `AES_STATUS_BUSY)
	   read(`AES_STATUS, status);

	 // output data
	 read(`AES_DOUT_0, aes_out[127:120]);
	 read(`AES_DOUT_1, aes_out[119:112]);
	 read(`AES_DOUT_2, aes_out[111:104]);
	 read(`AES_DOUT_3, aes_out[103:96]);
	 read(`AES_DOUT_4, aes_out[95:88]);
	 read(`AES_DOUT_5, aes_out[87:80]);
	 read(`AES_DOUT_6, aes_out[79:72]);
	 read(`AES_DOUT_7, aes_out[71:64]);
	 read(`AES_DOUT_8, aes_out[63:56]);
	 read(`AES_DOUT_9, aes_out[55:48]);
	 read(`AES_DOUT_10, aes_out[47:40]);
	 read(`AES_DOUT_11, aes_out[39:32]);
	 read(`AES_DOUT_12, aes_out[31:24]);
	 read(`AES_DOUT_13, aes_out[23:16]);
	 read(`AES_DOUT_14, aes_out[15:8]);
	 read(`AES_DOUT_15, aes_out[7:0]);

	 @(posedge CLK);
	 @(posedge CLK);
	 @(posedge CLK);
	 @(posedge CLK);

	 if (aes_out !== plain[i])
	   $display("ERROR: Expected %x Got %x", plain[i], aes_out);
	 else
	   $display("PASS : Expected %x Got %x", plain[i], aes_out);

	 iv_tmp = cipher[i];
	 
      end // for (i = 0; i < 4; i = i + 1)

      $stop;
   end // initial begin

endmodule // aes_h8_tb

